require 'morq/base'
require 'depot'		# QDBM

# - update label database (label -> lid /  lid -> mid... / mid -> lid...)

module Morq::Label
  class DB
    include Morq::Common

    def initialize(update=false)
      @db0 = open_morq_db("label0", update)
      @db1 = open_morq_db("label1", update)
      @db2 = open_morq_db("label2", update)
    end

    def close
      @db0.close
      @db1.close
      @db2.close
    end

    def add(mid, label)
      return if /^\w+$/ !~ label
      return if label == "archive" # do not label; just archive
      begin
        lid = @db0[label]
      rescue Depot::EANY
        lid = pack_id(@db0.size)
        @db0[label] = lid
      end
      mid = pack_id(mid)
      begin
        data = @db1[lid]
        return if list_include?(data, mid)
      rescue Depot::EANY
        # ignore
      end
      @db1.put(lid, mid, Depot::DCAT)
      @db2.put(mid, lid, Depot::DCAT)
    end

    def mid_data(label, n=0, unit=0)
      head = unit*n*4
      lid = @db0[label]
      data = @db1[lid] rescue ""
      if n > 0
        data = data[-head, unit*4] or ""
      end
      data
    end
    def get_labels(mid)
      labels = []
      hash = @db0.to_hash
      pmid = pack_id(mid)
      data = @db2[pmid] rescue ""
      list_each(data) do |lid|
        labels.push(hash.index(lid))
      end
      labels
    rescue
      return labels
    end

    def remove(mid, label)
      return if label == "archive" # do not label; just archive
      begin
        lid = @db0[label]
      rescue Depot::EANY
        # non existing label
        return
      end
      pmid = pack_id(mid)
      begin
        data = @db1[lid]
      rescue Depot::EANY
        # bad label
        return
      end
      list_delete(data, pmid)
      @db1.put(lid, data, Depot::DOVER)
      begin
        data = @db2[pmid]
      rescue Depot::EANY
        # inconsistent data
      end
      list_delete(data, lid)
      @db2.put(pmid, data, Depot::DOVER)
    end

    def kill(label)
      return if %w(inbox spam trash archive).include?(label)
      begin
        lid = @db0[label]
        @db0.delete(label)
      rescue Depot::EANY
        # non existing label
        return
      end
      begin
        data = @db1[lid]
        @db1.delete(lid)
      rescue Depot::EANY
        # bad label
        return
      end
      list_each(data) do |pmid|
        begin
          data = @db2[pmid]
          list_delete(data, lid)
          @db2.put(pmid, data, Depot::DOVER)
        rescue Depot::EANY
          # inconsistent data
        end
      end
    end

    def clear(label)
      begin
        lid = @db0[label]
      rescue Depot::EANY
        # non existing label
        return
      end
      begin
        data = @db1[lid]
        @db1.delete(lid)
      rescue Depot::EANY
        # bad label
        return
      end
      list_each(data) do |pmid|
        begin
          @db1.delete(pmid)
        rescue Depot::EANY
          # inconsistent data
        end
      end
    end

    def list
      keys = @db0.keys
      keys - ["archive"]
    end
  end

  def open_label_db(update=false)
    db = DB.new(update)
    begin
      yield db
    ensure
      db.close
    end
  end
end
