require 'morq/base'
require 'morq/label'
require 'morq/thread'

require 'iconv'

# LIST_INBOX:
# - LIST_LABEL(inbox)

# LIST_INBOX:
# - LIST_LABEL(inbox)

# LIST_LABEL(lid):
# - get mid... from label database (lid -> mid...)
# - LIST_INDEX(mid...)

# LIST_INDEX(mid...)
# - get index line from index database
# - sort by thread, date
# - list with thread indenting

module Morq::List
  include Morq::Common
  include Morq::Label
  include Morq::Thread

  @@encoding = "EUC-JP"
  @@extern_encoding = "EUC-JP"

  def set_encoding(str)
    @@encoding = str
#    Morq::Estraier.estencoding(str)
  end

  def self.encoding
    @@encoding
  end

  def set_extern_encoding(str)
    @@extern_encoding = str
  end

  def self.extern_encoding
    @@extern_encoding
  end

  class Box
    include Morq::Common

    def decode_b(str)
      str.unpack("m")[0]
    end

    def decode_q(str)
      str.unpack("M")[0]
    end

    def mime_decode(str)
      str.gsub!(/[\s\t]+/, " ")
      str.gsub!(/\?=\s+=\?/, "?==?")
      str.gsub(/=\?(.*?)\?(b|q)\?(.*?)\?=/i) do
        encode = $2.downcase
        charset = $1
        body = $3
        case encode
        when "b"
          body = decode_b(body)
        else
          body = decode_q(body)
        end
        Iconv.iconv(Morq::List.encoding, charset, body) rescue body
      end
    end

    def list_element(elem, n, out)
      time = Time.at(elem[4])
      from = mime_decode(elem[5]) 
      subj = mime_decode(elem[6])
      # mid flags pid tid yyyy-mm-dd hh:mm:ss th from subject
      out.printf "%d %d %d %d %s %d %s\t%s\n", elem[0], elem[1], elem[2], elem[3], time.strftime("%Y-%m-%d %H:%M:%S"), n, from, subj
      if elem[-1].size > 0
        for i in elem[-1]
          list_element(i, n+1, out)
        end
      end
    end

    def initialize(data, threading=true)
      list = []
      parents = Hash.new{[]}
      roots = {}
      open_morq_db("index") do |db|
        list_each(data) do |pmid|
          begin
            datum = unpack_index(db[pmid])
          rescue
            next
          end
          mid = unpack_id(pmid)
          datum.unshift mid
          datum.push datum[4]
          datum.push []
          parents[mid] = datum
          list << datum
        end
      end
      if threading
        list.sort!{|a,b| a[4]<=>b[4]} # sort by date
        list.delete_if do |datum|
          pid = datum[2]
          tid = datum[3]
          if pid != datum[0] && parents.key?(pid)
            parent = parents[pid]
            parent[-1].push datum
            if datum[4] > parent[-2]
              parent[-2] = datum[4]
            end
            true
          elsif roots.key?(tid)
            root = roots[tid]
            root[-1].push datum
            if datum[4] > root[-2]
              root[-2] = datum[4]
            end
            true
          else
            roots[tid] ||= datum
            false
          end
        end
        list.sort!{|a,b| a[-2]<=>b[-2]} # sort by latest date
      end
      @list = list
    end

    def list(out=STDOUT)
      for i in @list
        list_element(i, 0, out)
      end
    end

    def delete_elem(ary, mid)
      children = []
      ary.delete_if do |elem|
        if elem[0] == mid
          children.concat elem[-1]
          true
        else
          delete_elem(elem[-1], mid)
          false
        end
      end
      ary.concat children
    end

    def delete(mid)
      delete_elem(@list, mid)
    end
  end

  def box_label(label, n=0, unit=0)
    Box::new(open_label_db(){|db| db.mid_data(label, n, unit)})
  end

  def box_thread(tid)
    Box::new(get_thread_list(tid))
  end

  def box_query(query)
    Morq::List::Box.new(query_result(query), false)
  end

  def box_similarity(mid)
    raise RuntimeError, "Rast index does not support similarity search"
  end
end

if __FILE__ == $0
  include Morq::List

#  box = box_label("inbox")
  box = box_thread(1)
  box.list(STDOUT)
  box.delete(2)
  box.list(STDOUT)
end
