require 'morq/base'

# - update thread database (message-id -> mid / mid -> tid / tid -> mid...)

module Morq::Thread
  def open_msgid_db(&block)
    open_morq_db("msgid", true, &block)
  end

  def open_thread_db2
    # db1: mid -> tid, db2: tid -> mid...
    db1 = open_morq_db("thread1", true)
    db2 = open_morq_db("thread2", true)
    begin
      yield db1, db2
    ensure
      db1.close
      db2.close
    end
  end

  def register_thread(mid, headers)
    msgid = headers['message-id'].to_s
    return mid, mid unless msgid    # ignore mail without message-id
    pmid = pack_id(mid)
    pid = nil; tid = nil
    open_msgid_db() do |db0|
      db0.put(msgid, pmid, Depot::DOVER)
      if headers['in-reply-to'] && /<[^>]+>/ =~ headers['in-reply-to'].to_s
        pid = db0[$&] rescue nil
      end
      if pid.nil? and headers['references']
        articles = []
        headers['references'].to_s.scan(/<[^>]+>/) do |m|
          articles.push m
        end
        articles.reverse_each do |m|
          begin
            pid = db0[$&]
            break
          rescue
            # continue
          end
        end
      end

      open_thread_db2() do |db1, db2|
        if pid.nil?
          # no parent article (tid = self)
          tid = pid = pmid
        else
          tid = db1[pid] rescue pid
        end
        db1.put(pmid, tid, Depot::DOVER)
        db2.put(tid, pmid, Depot::DCAT)
      end
    end
    return unpack_id(pid), unpack_id(tid)
  end

  def get_thread_list(mid)
    ptid = pack_id(mid)
    open_thread_db2() do |db1, db2|
      tid = db1[ptid]
      list = db2[tid]
      return list
    end
  rescue Depot::EANY
    return ptid
  end
end
